from enum import Enum
from os import makedirs
import os.path as op
from colorama import Style, Fore
from datetime import datetime


def check_dir(*args, dir_exist_ok=True) -> str:
    path = op.abspath(op.join(*args))
    if op.isfile(path):
        path, _ = op.split(path)
        # raise IOError(f'{path} is a file instead of a directory.')
    if not dir_exist_ok and op.isdir(path):
        raise IOError(f'{path} already exist.')
    makedirs(path, exist_ok=True)
    return path


def datetime_str():
    return datetime.isoformat(datetime.now())[:-4]

class MessageIntent(Enum):
    DEFAULT=0
    INFO=1
    SUCCESS=2
    ERROR=3
    WARNING=4
    HIGHLIGHT=5

def cprint(*args, intent:MessageIntent=MessageIntent.DEFAULT):
    """ Print a (optionally colored) message with the date and time"""
    if intent == MessageIntent.INFO: clr = Fore.BLUE
    elif intent == MessageIntent.SUCCESS: clr = Fore.GREEN
    elif intent == MessageIntent.ERROR: clr = Fore.RED
    elif intent == MessageIntent.WARNING: clr = Fore.YELLOW
    elif intent == MessageIntent.HIGHLIGHT: clr = Fore.LIGHTMAGENTA_EX
    else: clr = Fore.WHITE
    print(f'{datetime_str()}:', clr, " ".join(str(a) for a in args), Style.RESET_ALL)


def merge_dicts(a: dict, b: dict) -> dict():
    """ Override a with b values 

        If a keys dot not existe in b, it is created
    """
    for key in b:
        if key in a and isinstance(a[key], dict) and isinstance(b[key], dict):
            a[key] = merge_dicts(a[key], b[key])
        else:
            a[key] = b[key]
    return a



import tempfile
import os.path as op

import yaml
from micropype.config import Config


@Config.register
class SubConfig(Config):
    name:   str = "Albert"
    age:    float

@Config.register
class AConfig(Config):
    num:        float = .3
    foo:        dict
    subject:    SubConfig


def test_config():
    # Create config from dict
    conf = {
        "paul": "jeje",
        "foo": {"item1": 0.1, "item2": 0.2},
        "subject": {
            "name": "ciceron",
            "age":  12
        }
    }

    config = AConfig(**conf)
    assert(config.num == .3)
    assert(config.subject.age == 12)

    # Export as yaml
    tmp_dir = tempfile.TemporaryDirectory()
    yaml_f = op.join(tmp_dir.name, "config.yaml")
    config.to_yaml(yaml_f)

    yml = yaml.safe_load(open(yaml_f, 'r'))
    assert(yml['num'] == .3)

    # Re-import from yaml
    yconfig = AConfig(yaml_f)
    assert(yconfig.foo['item2'] == config.foo['item2'])
    assert(yconfig.num == config.num)
    assert(yconfig.subject.age == config.subject.age)
    assert(yconfig.subject.name == config.subject.name)






from micropype.utils import merge_dicts


def test_dict_merge():
    a = {
        "a": 0,
        "b": {
            1: 200,
            2: 300
        }
    }

    b = {
        "b": {
            2: 400,
            4: 101
        }
    }

    c = merge_dicts(a, b)
    print(c)
    assert(c["a"] == 0)
    assert(c["b"][4] == 101)
    assert(c["b"][4] == 101)


from warnings import warn
import yaml
from micropype.utils import MessageIntent, cprint, merge_dicts


def read_annotations(cls):
    """ 
        Parameters
        ==========
    
    """
    attributes = {}
    for attr, attr_type in cls.__annotations__.items():
        default = cls.__getattribute__(cls, attr) if hasattr(cls, attr) else None
        t = attr_type.__name__
        attributes[attr] = (t, default)
    return attributes


class Config:
    _children = {}

    def __init__(self, yaml_f=None, priority="yaml", **kwargs) -> None:
        """
            Parameters
            ==========
            yaml_f: str
            priority: "yaml" or "kwargs"
                Which config values to use even if defined in the other method.
        """
        if yaml_f is not None:
            ycfg = yaml.safe_load(open(yaml_f, 'r'))
            if priority == "yaml":
                kwargs = merge_dicts(kwargs, ycfg)
            elif priority == "kwargs":
                kwargs = merge_dicts(ycfg, kwargs)
            else:
                raise ValueError("")

        attributes = read_annotations(self.__class__)

        used_kwargs = 0 
        for attr_name, (attr_t, default_v) in attributes.items():
            useDefault = not attr_name in kwargs
            if not useDefault:
                used_kwargs += 1

            if attr_t in self._children:
                value = self._children[attr_t] if useDefault else self._children[attr_t](**kwargs[attr_name])
            else:
                value = default_v if useDefault else kwargs[attr_name]
            self.__setattr__(attr_name, value)
        if used_kwargs < len(kwargs.keys()):
            for k in kwargs.keys():
                if not k in attributes.keys():
                    cprint(
                        f'"{k}" is not an attribute of {self.__class__.__name__}. Skipping this setting.',
                        intent=MessageIntent.WARNING
                    )
                    continue

    @classmethod
    def register(cls, child_class):
        cls._children[child_class.__name__] = child_class
        return child_class
    

    def to_dict(self) -> dict:
        result = {}
        for name, value in self.__dict__.items():
            if isinstance(value, Config):
                result[name] = value.to_dict()
            elif isinstance(value, list):
                result[name] = [item.to_dict() if isinstance(item, Config) else item for item in value]
            elif isinstance(value, dict):
                result[name] = {key: value.to_dict() if isinstance(value, Config) else value for key, value in value.items()}
            else:
                result[name] = value
        return result

    def to_yaml(self, filepath:str):
        with open(filepath, 'w') as fp:
            yaml.dump(self.to_dict(), fp)


# if __name__ == "__main__":
#     class SubConfig(Config):
#         name:   str = "Albert"
#         age:    float
#     class AConfig(Config):
#         num:        float = .3
#         foo:        dict
#         subject:    SubConfig

#     conf = {
#         "paul": "jeje",
#         "foo": {"item1": 0.1, "item2": 0.2},
#         "subject": {
#             "name": "ciceron",
#             "age":  12
#         }
#     }

#     config = AConfig(**conf)
#     pass




import tempfile
import os.path as op
from micropype import cached_run, cached_function_call
from micropype.config import Config


def add(x, y, out_f):
    with open(out_f, 'w') as fp:
        fp.write(f"{x+y}")


def test_pipeline():
    tmp_dir = tempfile.TemporaryDirectory()
    
    log_f = op.join(tmp_dir.name, "test.log")
    out_f = op.join(tmp_dir.name, "lsOutput.txt")
    cached_run(
        f"ls -lrt >> {out_f}",
        out_f,
        "Listing current folder items",
        log_f
    )
    with open(out_f, 'r') as fp:
        assert(fp.readline().startswith("total"))

    out_f = op.join(tmp_dir.name, "add.txt")
    cached_function_call(
        add,
        [2, 4.0, out_f],
        [out_f],
        title=f"Addition",
        log=log_f
    )
    with open(out_f, 'r') as fp:
        res = float(fp.readline())
        assert(res == 6)



def test_pipeline_with_config():
    class PipeConfig(Config):
        do_step1: bool = True
        do_step2: bool = True

    cfg = PipeConfig(
        do_step1=True, 
        do_step2=False
    )

    tmp_dir = tempfile.TemporaryDirectory()
    
    log_f = op.join(tmp_dir.name, "test.log")
    out_f = op.join(tmp_dir.name, "lsOutput.txt")
    if cfg.do_step1:
        cached_run(
            f"ls -lrt >> {out_f}",
            out_f,
            "Listing current folder items",
            log_f
        )
    with open(out_f, 'r') as fp:
        assert(fp.readline().startswith("total"))

    out_f = op.join(tmp_dir.name, "add.txt")
    if cfg.do_step2:
        cached_function_call(
            add,
            [2, 4.0, out_f],
            [out_f],
            title=f"Addition",
            log=log_f
        )
    assert(op.isfile(out_f) == False)



import subprocess
import os.path as op
from time import time
from datetime import datetime
from typing import Iterable, List, Union
from .utils import MessageIntent, cprint
import sys
import traceback


def write_log(log_f, *text: Union[str, List[str]]):
    """ Append text to the log text file. """
    # text = [text] if isinstance(text, str) else text
    with open(log_f, 'a+') as fp:
        fp.writelines(text)

def get_versions():
    return {
        "python": f"{sys.version[0]}.{sys.version[1]}"
    }


def versions_2_txt(versions:dict=None):
    txt = "### Versions:\n"
    if versions is not None:
        for k, v in versions.items():
            txt += f"{k}: {v}\n"
    txt += "#############\n"
    return txt
    

def run_cmd(cmd, title=None, log=None, versions:dict=None, raise_errors=True):
    """ Run a command in a sub process """
    splitted_cmd = cmd #cmd.split(' ')

    if title:
        cprint(f"Start {title}", intent=MessageIntent.INFO)
    cprint(cmd)
    tic = time()
    try:
        if log:
            write_log(log, 
                "\n",
                "#" * 80,
                f"\n####### {title} ######\n",
                versions_2_txt(versions),
                cmd,
                f"\n##########\nStarted at: {datetime.isoformat(datetime.now())}"
            )
            tic = time()
            output = subprocess.check_output(splitted_cmd, stderr=subprocess.STDOUT, shell=True)
            toc = time()
            write_log(log,
                output.decode("utf-8"),
                f"\nFinshed at: {datetime.isoformat(datetime.now())} - tooks {toc-tic:.03f}s",
                "\n#####\n\n\n"
            )
        else:
            subprocess.run(splitted_cmd, shell=True)
    except Exception as e:
        tb = traceback.format_exc()
        cprint(f"An error occured while running: {cmd}", intent=MessageIntent.ERROR)
        if log:
            write_log(log,
                f"Errored at {datetime.isoformat(datetime.now())}:\n",
                str(e), "\n#####\n\n\n"
            )
        if raise_errors:
            raise e
        else:
            cprint(str(e), intent=MessageIntent.ERROR)
            cprint(tb, intent=MessageIntent.ERROR)
            return -1
    if log:
        cprint(f'Console outputs saved to:', log, intent=MessageIntent.INFO)
    if title:
        cprint(f"{title} tooks {time()-tic:.02f}s", intent=MessageIntent.SUCCESS)
    else:
        cprint(f"done in {time()-tic:.02f}s", intent=MessageIntent.SUCCESS)
    return 0


def cached_run(cmd, out_files:List[str]=None, title=None, log=None, 
               versions:dict=None, raise_errors=True, verbose=False):
    """ Run the command only if one of the out_files is missing """
    if isinstance(out_files, str):
        out_files = [out_files]

    do_run = False
    if out_files is not None:
        for f in out_files:
            if not op.isfile(f) and not op.isdir(f):
                do_run = True
                break
    
    if do_run:
        return run_cmd(cmd, title, log, versions, raise_errors)
    if verbose:
        cprint('Using cached files for', title if title else cmd, intent=MessageIntent.WARNING)


def _run_func(func, args):
    if args is None:
        func()
    elif isinstance(args, dict):
        func(**args)
    elif isinstance(args, Iterable):
        func(*args)
    else:
        raise ValueError("arguments hould be either None, a list or a dict.")
 

def run_func(func, args, title=None, log=None, versions:dict=None, raise_errors=True):
    """ Run a command in a sub process """
    if title:
        cprint(f"Start {title}", intent=MessageIntent.INFO)
    cprint(func.__name__)
    cprint('args:\n')
    for arg in args:
        cprint('\t', arg)

    tic = time()
    try:
        if log:
            write_log(log, 
                "\n",
                "#" * 80,
                f"\n####### {title} ######\n",
                versions_2_txt(versions),
                func.__name__ if hasattr(func, '__name__') else func,
                f"\n##########\nStarted at: {datetime.isoformat(datetime.now())}"
            )
            tic = time()
            _run_func(func, args)
            toc = time()
            write_log(log,
                f"\nFinshed at: {datetime.isoformat(datetime.now())} - tooks {toc-tic:.03f}s",
                "\n#####\n\n\n"
            )
        else:
            _run_func(func, args)
    except Exception as e:
        tb = traceback.format_exc()
        cprint(f"An error occured while running: {func.__name__}", intent=MessageIntent.ERROR)
        cprint(str(e), intent=MessageIntent.ERROR)
        if log:
            write_log(log,
                f"Errored at {datetime.isoformat(datetime.now())}:\n",
                str(e), "\n#####\n\n\n"
            )
        if raise_errors:
            raise e
        else:
            cprint(str(e), intent=MessageIntent.ERROR)
            cprint(tb, intent=MessageIntent.ERROR)
            return -1
    if log:
        cprint(f'Console outputs saved to:', log, intent=MessageIntent.INFO)
    if title:
        cprint(f"{title} tooks {time()-tic:.02f}s", intent=MessageIntent.SUCCESS)
    else:
        cprint(f"done in {time()-tic:.02f}s", intent=MessageIntent.SUCCESS)
    return 0


def cached_function_call(func, args, out_files:List[str]=None, title=None, log=None, 
                         versions:dict=None, raise_errors=True, verbose=False):
    """ Run python function only if one of the out_files is missing """
    if isinstance(out_files, str):
        out_files = [out_files]

    do_run = False
    if  out_files is not None:
        for f in out_files:
            if not op.isfile(f) and not op.isdir(f):
                do_run = True
                break
    
    if do_run:
        return run_func(func, args, title, log, versions, raise_errors)
    if verbose:
        cprint('Using cached files for', title if title else func.__name__, intent=MessageIntent.WARNING)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from setuptools import find_packages, setup


setup(
    name="micropype",
    version="0.2.1",
    packages=find_packages(),
    author="Bastien Cagna",
    description="Very basic pipelining toolbox",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    license='BSD 3',
    entry_points = {},
    install_requires=["pyyaml", "colorama"],
    extras_require={ "dev": [ "pytest" ]},
    include_package_data=True,
)
